#include <iostream>
#include <blt/std/allocator.h>
#include <blt/gfx/window.h>
#include <blt/gfx/state.h>
#include <blt/gfx/stb/stb_image.h>
#include <blt/std/hashmap.h>
#include <functions.h>
#include "blt/gfx/renderer/resource_manager.h"
#include "blt/gfx/renderer/batch_2d_renderer.h"
#include "blt/std/assert.h"
#include "imgui.h"
#include <variant>
#include <random>
#include <queue>
#include <stack>

blt::gfx::matrix_state_manager global_matrices;
blt::gfx::resource_manager resources;
blt::gfx::batch_renderer_2d renderer_2d(resources);

constexpr blt::i32 MAX_DEPTH = 17;

struct node;

blt::area_allocator<node, 32000> node_allocator;

node* createNode(function_t type)
{
    auto* n = node_allocator.allocate(1);
    node_allocator.construct(n, type);
    return n;
}


void destroyNode(node* n)
{
    node_allocator.destroy(n);
    node_allocator.deallocate(n, 1);
}

class tree;

struct node
{
        friend tree;
    private:
        // only valid up to the argc
        std::array<node*, MAX_ARGS> sub_nodes{nullptr};
        size_t argc = 0;
        
        function_t type = function_t::ADD;
        data_t data{};
        
        std::optional<image> img;
        
        void grow(bool use_terminals)
        {
            auto min_children = function_arg_min_map[to_underlying(type)];
            auto max_children = function_arg_max_map[to_underlying(type)];
            
            static thread_local std::random_device dev;
            static thread_local std::mt19937_64 engine{dev()};
            std::uniform_int_distribution<blt::i32> dist(min_children, max_children);
            static std::uniform_int_distribution<int> choice(0, 1);
            
            argc = dist(engine);
            if (argc == 0)
                return;
            const auto& allowed_args_args = function_arg_allowed_map[to_underlying(type)];
            for (size_t i = 0; i < argc; i++)
            {
                // 50/50 chance to either use a terminal or use from the function list.
                populate_node(i, engine, allowed_args_args[i], choice(engine) || use_terminals);
            }
        }
        
        void full(bool use_terminals)
        {
            static thread_local std::random_device dev;
            static thread_local std::mt19937_64 engine{dev()};
            
            argc = function_arg_max_map[to_underlying(type)];
            if (argc == 0)
                return;
            
            const auto& allowed_args_args = function_arg_allowed_map[to_underlying(type)];
            for (size_t i = 0; i < argc; i++)
                populate_node(i, engine, allowed_args_args[i], use_terminals);
        }
        
        inline void populate_node(size_t i, std::mt19937_64& engine, const allowed_funcs<function_t>& allowed_args, bool use_terminal)
        {
            if (use_terminal)
            {
                auto terminals = intersection(allowed_args, FUNC_ALLOW_TERMINALS_SET);
                if (terminals.empty())
                {
                    terminals = FUNC_ALLOW_TERMINALS;
//                    BLT_INFO("%s:", function_name_map[to_underlying(type)].c_str());
//                    for (auto v : allowed_args)
//                        BLT_INFO(function_name_map[to_underlying(v)]);
                }
                std::uniform_int_distribution<blt::size_t> select(0, terminals.size() - 1);
                sub_nodes[i] = createNode(terminals[select(engine)]);
            } else
            {
                auto non_terminals = intersection_comp(allowed_args, FUNC_ALLOW_TERMINALS_SET);
                if (non_terminals.empty())
                {
                    BLT_WARN("Empty non-terminals set! Filling from terminals!");
                    std::uniform_int_distribution<blt::size_t> select(0, allowed_args.size() - 1);
                    sub_nodes[i] = createNode(allowed_args[select(engine)]);
                } else
                {
                    std::uniform_int_distribution<blt::size_t> select(0, non_terminals.size() - 1);
                    sub_nodes[i] = createNode(non_terminals[select(engine)]);
                }
            }
        }
        
        void print_tree()
        {
            if (argc > 0)
                std::cout << "(";
            if (argc > 0)
                std::cout << function_name_map[to_underlying(type)] << " ";
            else
            {
                if (type == function_t::SCALAR)
                {
                    evaluate();
                    std::cout << img->get().x() << " ";
                } else if (type == function_t::COLOR)
                {
                    evaluate();
                    std::cout << '{' << img->get().x() << ", " << img->get().y() << ", " << img->get().z() << "} ";
                } else
                    std::cout << function_name_map[to_underlying(type)] << " ";
            }
            for (size_t i = 0; i < argc; i++)
                sub_nodes[i]->print_tree();
            if (argc > 0)
                std::cout << ") ";
        }
    
    public:
        explicit node(function_t type): type(type)
        {
            static thread_local std::random_device dev;
            static thread_local std::mt19937_64 engine{dev()};
            static thread_local std::uniform_real_distribution<float> dist(0.0f, 1.0f);
            if (type == function_t::SCALAR)
                data[0] = dist(engine);
            else if (type == function_t::COLOR)
            {
                for (auto& v : data)
                    v = dist(engine);
            }
        }
        
        static node* construct_random_tree()
        {
            static std::random_device dev;
            static std::mt19937_64 engine{dev()};
            std::uniform_int_distribution<int> choice(0, 1);
            
            static auto NON_TERMINALS = intersection_comp(FUNC_ALLOW_ANY, FUNC_ALLOW_TERMINALS_SET);
            
            std::uniform_int_distribution<int> select(0, NON_TERMINALS.size() - 1);
            
            node* n = createNode(NON_TERMINALS[select(engine)]);
            
            std::queue<std::pair<node*, size_t>> grow_queue;
            size_t current_depth = 0;
            grow_queue.emplace(n, current_depth);
            while (!grow_queue.empty())
            {
                auto front = grow_queue.front();
                if (front.second != current_depth)
                    current_depth++;
                
                if (choice(engine))
                    front.first->grow(front.second >= MAX_DEPTH);
                else
                    front.first->full(front.second >= MAX_DEPTH);
                
                for (size_t i = 0; i < front.first->argc; i++)
                    grow_queue.emplace(front.first->sub_nodes[i], current_depth + 1);
                grow_queue.pop();
            }
            
            return n;
        }
        
        void evaluate()
        {
            img = image{};
            std::array<image*, MAX_ARGS> sub_node_images{nullptr};
            for (size_t i = 0; i < argc; i++)
            {
                BLT_ASSERT(sub_nodes[i] != nullptr && sub_nodes[i]->img.has_value() && "Node must have evaluated children!");
                sub_node_images[i] = &sub_nodes[i]->img.value();
            }
#define FUNC_DEFINE(NAME, MIN_ARGS, MAX_ARGS, FUNC, ...) case function_t::NAME: {                                                                               \
                if (FUNC_ALLOW_TERMINALS_SET.contains(function_t::NAME)){                                                                                       \
                    FUNC(img.value(), 0, 0, argc, const_cast<const image**>(sub_node_images.data()), data);                                                     \
                } else {                                                                                                                                        \
                    for (blt::i32 y = 0; y < height; y++) {                                                                                                     \
                        for (blt::i32 x = 0; x < width; x++) {                                                                                                  \
                            FUNC(img.value(), static_cast<float>(x), static_cast<float>(y), argc, const_cast<const image**>(sub_node_images.data()), data);     \
                        }                                                                                                                                       \
                    }                                                                                                                                           \
                }                                                                                                                                               \
            }                                                                                                                                                   \
            break;
            
            switch (type)
            {
                FUNC_FUNCTIONS
                default:
                    BLT_WARN("How did we get here?");
                    break;
            }
#undef FUNC_DEFINE
        }
        
        void evaluate_tree()
        {
            for (size_t i = 0; i < argc; i++)
                sub_nodes[i]->evaluate_tree();
            evaluate();
        }
        
        void printTree()
        {
            print_tree();
            std::cout << std::endl;
        }
        
        bool hasImage()
        {
            return img.has_value();
        }
        
        image& getImage()
        {
            return img.value();
        }
        
        ~node()
        {
            for (auto* node : sub_nodes)
                destroyNode(node);
        }
};

class tree
{
    public:
        node* root;
    
    
};

node* root;
blt::gfx::texture_gl2D* texture;

bool fucky(float f)
{
    return f == std::numeric_limits<float>::quiet_NaN() || f == -std::numeric_limits<float>::quiet_NaN() ||
           f == std::numeric_limits<float>::infinity() ||
           f == -std::numeric_limits<float>::infinity();
}

void init()
{
    global_matrices.create_internals();
    renderer_2d.create();
    
    texture = new blt::gfx::texture_gl2D(width, height);
    resources.set("img", texture);
    resources.enqueue("../libraries/BLT-With-Graphics-Template/resources/textures/parkerfemBOY.png", "cum");
    
    resources.load_resources();
}

void update(std::int32_t w, std::int32_t h)
{
    global_matrices.update_perspectives(w, h, 90, 0.1, 2000);
    glDisable(GL_BLEND);
    
    if (ImGui::Button("Regenerate"))
    {
        BLT_INFO("Regen tree");
        destroyNode(root);
        root = node::construct_random_tree();
        root->evaluate_tree();
        BLT_INFO("Preprocess");
        
        float mx = 0, my = 0, mz = 0;
        float sx = 0, sy = 0, sz = 0;
        if (root->hasImage())
        {
            for (auto& v : root->getImage().getData())
            {
                //v = v.normalize();
                for (int i = 0; i < 3; i++)
                    v[i] = std::abs(v[i]);
                mx = std::max(v.x(), mx);
                my = std::max(v.y(), my);
                mz = std::max(v.z(), mz);
                sx = std::min(v.x(), sx);
                sy = std::min(v.y(), sy);
                sz = std::min(v.z(), sz);
            }
            for (auto& v : root->getImage().getData())
            {
                if (mx - sx != 0)
                    v[0] = (v.x() - sx) / (mx - sx);
                if (my - sy != 0)
                    v[1] = (v.y() - sy) / (my - sy);
                if (mz - sz != 0)
                    v[2] = (v.z() - sz) / (mz - sz);
            }
        }
        
        
        BLT_INFO("Uploading");
        //delete texture;
        if (root->hasImage())
            texture->upload((void*) root->getImage().getData().data(), GL_RGB, 0, 0, 0, -1, -1, GL_FLOAT);
        //texture->upload((void*) test.data(), GL_RGBA, 0, 0, 0, width, height, GL_UNSIGNED_BYTE);
        //texture->upload(file);
        //texture = new blt::gfx::texture_gl2D(file);
        resources.set("img", texture);
    }
    
    if (ImGui::Button("Display"))
    {
        if (root)
            root->printTree();
    }
    
    if (ImGui::Button("Eval"))
    {
        if (root && root->hasImage())
            BLT_DEBUG(eval_AMF(root->getImage()));
    }
    
    auto lw = 512.0f;
    auto lh = 512.0f;
    //renderer_2d.drawRectangle(blt::vec4{0.5, 0.0, 1.0, 1.0},
    //                          {static_cast<float>(w) / 2.0f, static_cast<float>(h) / 2.0f, static_cast<float>(w), static_cast<float>(h)});
    renderer_2d.drawRectangle("img", {static_cast<float>(w) / 2.0f, static_cast<float>(h) / 2.0f, lw, lh});
    
    global_matrices.update();
    renderer_2d.render();
}

int main()
{
    blt::gfx::init(blt::gfx::window_data{"Window of GP test", init, update}.setSyncInterval(1));
    global_matrices.cleanup();
    resources.cleanup();
    renderer_2d.cleanup();
    blt::gfx::cleanup();
    return 0;
}
